/******************************************************************************/
/* FB database user substitution                                              */
/* Script privprocs.sql, created Dec 21, 2011 by Vitezslav Svejdar            */
/* Revised:                                                                   */
/*   Jul 19, 2013, cosmetic changes                                           */
/*   Jul 24, 2013 V 1.3:                                                      */
/*     Better explanation in comments, improved detection of relevant         */
/*     grantors in Step 2                                                     */
/*   Jul 31, 2013 V 1.4:                                                      */
/*     Improved treatment of cases when SYSDBA or DBO grant something         */
/*     on an object that they do not own                                      */
/*   Aug 8, 2013 V 2.0:                                                       */
/*     Improved logging: procedures rewritten as select procedures            */
/*     New treatment of SYSDBA and db owner                                   */
/******************************************************************************/
/*
We consider two database users, :DUSER who has died, and :MUSER, a new master,
who should be granted all the :DUSER's privileges, and assumes the responsibility
to grant all privileges that :DUSER granted. These two users must already be
defined in the security database of your Firebird installation.

The process is based on Thomas Steinmaurer's script and his Delphi utility, but
contains some refinements and is more general: it does not care who of the two
users, if any, is the database owner. If :DUSER is the db owner then
the ownership is transferred to :MUSER.

:MUSER can be anybody except 'PUBLIC'.
:DUSER should be one of the users given by:
select distinct rdb$owner_name from rdb$relations
union select distinct rdb$owner_name from rdb$procedures
union select distinct rdb$owner_name from rdb$roles
union select distinct rdb$grantor from rdb$user_privileges
union select distinct rdb$user from rdb$user_privileges where rdb$user_type=8;
but not :MUSER and not 'PUBLIC'.

Step 1, preperation and consistency check
DUSER connects to database and runs this script to create two publicly
executable procedures REDIR$$PRIV and GRANT$$PRIV (below), and to recreate
the X privilege on every procedure he owns. This latter step is done for the case
that ownership of procedures or privileges on them is corrupted.
Then within the same connection, DUSER runs the select statement
in the script invalidpriv.sql to detect invalid privileges.
Invalid privileges are probably caused by direct updates of system tables.
Some of them could be harmless for the substitution process.
The select statement should return empty dataset.

Step 2, finding relevant grantors
Still within the same connection, the following select statement yields the list
of all relevant grantors, where a relevant grantor is almost the same as a user
who grants privileges to DUSER. The first two (constant) selects are obvious,
the third select statement yields all users who grant anything to DUSER.
The last select statement returns one line containing db owner name in case
that DUSER owns any role and none of DUSER and MUSER is 'SYSDBA'; otherwise it
contributes nothing to the list of relevant grantors:
select cast(:duser as char(31)) from rdb$database
union select cast(:muser as char(31)) from rdb$database
union select distinct rdb$grantor from rdb$user_privileges where rdb$user_type=8 and rdb$user=:duser
union select distinct t.rdb$owner_name from rdb$relations t join rdb$roles r on 0=0
where t.rdb$relation_name='RDB$ROLES' and r.rdb$owner_name=:duser
and :duser<>'SYSDBA' and :muser<>'SYSDBA';

Step 3, granting privileges to MUSER
Each user detected in Step 2, but DUSER first, connects and says:
select * from REDIR$$PRIV(:duser,:muser).
A part of DUSER's execution is redirecting his ownership of tables and procedures
to MUSER. If DUSER owns any roles, then part of db owner's execution is
redirecting the ownership of these roles to MUSER.
For each privilege that any user A<>DUSER granted for DUSER, and in some cases
also if A=DUSER, this step creates a copy of that privilege granted to MUSER.
The troublesome cases are the following.
(a) DUSER='SYSDBA' granted X to himself on a procedure that neither he nor MUSER
owns. In this case we believe that MUSER, as an inheritor of DUSER, should be
granted that privilege. SYSDBA is the only user, who can grant it, and thus he
does not disappear without any trace.
(b) DUSER being a special user (SYSDBA in case of tables, SYSDBA or db owner
in case of roles) granted something on a table or a role without having
grant option (admin option).

Step 4, responsibilities of MUSER
MUSER connects and says select * from GRANT$$PRIV(:duser). For any privilege
granted by DUSER a copy granted by MUSER is thus created.
drop procedure REDIR$$PRIV;
drop procedure GRANT$$PRIV;

Step 5, cleaning privileges granted by DUSER
DUSER connects and says:
delete from rdb$user_privileges g where g.rdb$grantor=current_user
and (g.rdb$object_type<>5 or current_user<>'SYSDBA' or :muser=
(select p.rdb$owner_name from rdb$procedures p where p.rdb$procedure_name=g.rdb$relation_name))

Step 6, cleaning privileges granted to DUSER
Each user detected in Step 2 except DUSER, but including MUSER, connects and says
delete from rdb$user_privileges where rdb$grantor=current_user
and rdb$user=:duser and rdb$user_type=8

Step 7, backup and restore under the db owner. This step is not essential,
the database is in consistent state already.
*/
/******************************************************************************/
/****                     Create Three Stored Procedures                   ****/
/******************************************************************************/
set term ^ ;

create or  alter procedure redir$$priv (DUSER char(31),MUSER char(31))
returns (GSTMT varchar(240) character set none)
as
declare variable PRIV char(6) character set NONE;
declare variable GOPTION integer = 1;
declare variable RNAME char(31) character set NONE;
declare variable FNAME char(31) character set NONE;
declare variable DBOWNER char(31) character set NONE;
begin
/* Current_user reviews all privileges that he granted to DUSER,              */
/* and grants their copy to  MUSER.                                           */
  if (duser is null) then goption=goption/0;
  if (muser is null) then goption=goption/0;
  if (duser=muser) then goption=goption/0;
  if (duser='PUBLIC' or muser='PUBLIC') then goption=goption/0;
  select rdb$owner_name from rdb$relations where rdb$relation_name='RDB$DATABASE' into :dbowner;
/* Rights on tables */
  /* If the user who is substituted out is SYSDBA, he might have granted something without */
  /* having grant option. In this case we give him the grant option. That will later */
  /* equip MUSER with sufficient rights. */
  if (current_user=duser and current_user='SYSDBA') then
  for select distinct rdb$privilege,rdb$relation_name,rdb$field_name
  from rdb$user_privileges where rdb$grantor=current_user and rdb$object_type=0
  into :priv,:rname,:fname do
  begin
    if (priv in ('U','D')) then -- It seems that Firebird does not like U or D without S
    begin
      gstmt='grant select on "'||rname||'" to '||current_user||' with grant option';
      suspend;
      execute statement gstmt;
    end
    if (priv='A') then gstmt='all';
    if (priv='S') then gstmt='select';
    if (priv='D') then gstmt='delete';
    if (priv='I') then gstmt='insert';
    if (priv='R') then gstmt='references';
    if (priv='U') then gstmt='update';
    if (priv='R' or priv='U') then if (not fname is null) then gstmt=gstmt||' ("'||fname||'")';
    gstmt='grant '||gstmt||' on "'||rname||'" to '||current_user||' with grant option';
    suspend;
    execute statement gstmt;
  end
  -- Now the essential step for all users
  for select rdb$privilege,rdb$grant_option,rdb$relation_name,rdb$field_name
  from rdb$user_privileges where rdb$grantor=current_user and rdb$user=:duser
  -- rdb$object_type specifies type of object on which a privilege has
  -- been granted: 0 table or view, 5 procedure, 13 role.
  and rdb$object_type=0 -- Rights on tables and views
  -- rdb$user_type specifies type of "user" to whom a privilege has been granted:
  -- 1 view, 2 trigger, 5 procedure, 8 user, 13 role. It is probably possible that
  -- a procedure, for example, has the same name as a user.
  and rdb$user_type=8 -- So we have to distinguish privileges to users
  and (rdb$grantor<>:duser or rdb$grant_option=1)
  into :priv,:goption,:rname,:fname do
  begin
    if (priv='A') then gstmt='all';
    if (priv='S') then gstmt='select';
    if (priv='D') then gstmt='delete';
    if (priv='I') then gstmt='insert';
    if (priv='R') then gstmt='references';
    if (priv='U') then gstmt='update';
    if (priv='R' or priv='U') then if (not fname is null) then gstmt=gstmt||' ("'||fname||'")';
    gstmt='grant '||gstmt||' on "'||rname||'" to '||muser;
    if (goption=1) then gstmt=gstmt||' with grant option';
    suspend;
    execute statement gstmt;
  end
/* Rights on procedures. The important cases are when A<>DUSER granted X to DUSER, */
  if (current_user<>duser) then for select rdb$relation_name
    from rdb$user_privileges where rdb$grantor=current_user and rdb$user=:duser
    and rdb$object_type=5 and rdb$user_type=8
    into :rname do
    begin
      gstmt='grant execute on procedure "'||rname||'" to '||muser;
      suspend;
      execute statement 'grant execute on procedure "'||rname||'" to '||muser;
    end
  else -- and when SYSDBA=DUSER granted X on a proc that neither he nor MUSER owns
    if (current_user='SYSDBA') then for select g.rdb$relation_name
    from rdb$user_privileges g join rdb$procedures p on g.rdb$relation_name=p.rdb$procedure_name
    where g.rdb$object_type=5 and g.rdb$grantor=:duser and g.rdb$user=:duser and g.rdb$user_type=8
    and p.rdb$owner_name<>:duser and p.rdb$owner_name<>:muser into :rname do
    begin
      gstmt='grant execute on procedure "'||rname||'" to '||muser;
      suspend;
      execute statement 'grant execute on procedure "'||rname||'" to '||muser;
    end
/* Rights on roles */
  if (current_user=duser and (current_user='SYSDBA' or current_user=dbowner)) then
  for select distinct rdb$relation_name
  from rdb$user_privileges where rdb$grantor=current_user and rdb$object_type=13
  into :rname do
  begin
    gstmt='grant "'||rname||'" to '||current_user||' with admin option';
    suspend;
    execute statement gstmt;
  end
  for select rdb$relation_name,rdb$grant_option
  from rdb$user_privileges where rdb$grantor=current_user and rdb$user=:duser
  and rdb$object_type=13 and rdb$user_type=8
  and (rdb$grantor<>:duser or rdb$grant_option=2)
  into :rname,:goption do
  begin
    gstmt='grant "'||rname||'" to '||muser;
    if (goption=2) then gstmt=gstmt||' with admin option';
    suspend;
    execute statement gstmt;
  end
/* Ownership changes */
  if (current_user=duser) then -- Tables and procedures
  begin
    update rdb$relations set rdb$owner_name=:muser where rdb$owner_name=current_user;
    update rdb$procedures set rdb$owner_name=:muser where rdb$owner_name=current_user;
  end
  if (current_user='SYSDBA' or current_user=dbowner) then -- Roles. Experiment shows that we need execute statement, not just update.(?)
  begin
    gstmt='update rdb$roles set rdb$owner_name='''||muser||''' where rdb$owner_name='''||duser||'''';
    suspend;
    execute statement gstmt;
  end
end^

create or alter procedure grant$$priv (DUSER char(31))
returns (GSTMT varchar(240) character set none)
as
declare variable AUSER char(31) character set NONE;
declare variable PRIV char(6) character set NONE;
declare variable GOPTION integer = 1;
declare variable RNAME char(31) character set NONE;
declare variable FNAME char(31) character set NONE;
declare variable USERTYPE integer;
declare variable OBJTYPE integer;
begin
/* MUSER reviews all privileges that were granted by DUSER,                   */
/* and grants their copy to the same users/objects.                           */
  if (duser is null) then goption=goption/0;
  for select g.rdb$user,g.rdb$privilege,g.rdb$grant_option,g.rdb$relation_name,
    g.rdb$field_name,g.rdb$user_type,g.rdb$object_type
  from rdb$user_privileges g left join rdb$procedures p on g.rdb$relation_name=p.rdb$procedure_name
  where g.rdb$grantor=:duser and (g.rdb$object_type in (0,13) or p.rdb$owner_name=current_user)
  into :auser,:priv,:goption,:rname,:fname,:usertype,:objtype do
  begin
    if (usertype=8 and auser=duser) then auser=current_user;
    if (priv='A') then gstmt='all on'; -- I doubt that this case is at all possible
    if (priv='S') then gstmt='select on';
    if (priv='D') then gstmt='delete on';
    if (priv='I') then gstmt='insert on';
    if (priv='R') then gstmt='references';
    if (priv='U') then gstmt='update';
    if (priv in ('R','U')) then if (not fname is null) then gstmt=gstmt||' ("'||fname||'") on'; else gstmt=gstmt||' on';
    if (priv='X') then gstmt='execute on procedure';
    if (priv='M') then gstmt='';
    gstmt='grant '||gstmt||' "'||rname||'" to ';
    if (usertype=1) then gstmt=gstmt||'view ';
    if (usertype=2) then gstmt=gstmt||'trigger ';
    if (usertype=5) then gstmt=gstmt||'procedure ';
    -- The syntax of grant statement does not distinguish between granting something
    -- to user and granting to role. So we have no special treatment of the case where
    -- usertype=13; it is treated like if usertype were 8. I do not know what happens
    -- if some role has the same name as some user.
    gstmt=gstmt||auser;
    if (goption=1 and objtype=0) then gstmt=gstmt||' with grant option';
    if (goption=2 and priv='M') then gstmt=gstmt||' with admin option';
    suspend;
    execute statement gstmt;
  end
--  for select rdb$procedure_name from rdb$procedures where rdb$owner_name=current_user into :rname do
--    execute statement 'grant execute on procedure "'||rname||'" to '||current_user;
end^

create or alter procedure recreate$$xpriv 
as
declare variable RNAME char(31) character set NONE;
begin
  for select rdb$procedure_name from rdb$procedures where rdb$owner_name=current_user into :rname do
    execute statement 'grant execute on procedure "'||rname||'" to '||current_user;
end^

set term ; ^

/******************************************************************************/
/*  One of the procedures is immediately executed by DUSER,                   */
/*  the two remaining will be executed by all grantors, including DUSER       */
/******************************************************************************/
execute procedure recreate$$xpriv;
drop procedure recreate$$xpriv;
grant execute on procedure redir$$priv to public;
grant execute on procedure grant$$priv to public;
